@@ -25,7 +25,7 @@ const std::string galera::ReplicatorSMM::Param::key_format =
 const std::string galera::ReplicatorSMM::Param::max_write_set_size =
     common_prefix + "max_ws_size";
 
-int const galera::ReplicatorSMM::MAX_PROTO_VER(9);
+int const galera::ReplicatorSMM::MAX_PROTO_VER(10);
 
 galera::ReplicatorSMM::Defaults::Defaults() : map_()
 {
@@ -124,7 +124,7 @@ galera::ReplicatorSMM::ParseOptions::ParseOptions(Replicator&       repl,
                                                   gu::Config&       conf,
                                                   const char* const opts)
 {
-    conf.parse(opts);
+    if (opts) conf.parse(opts);
 
     if (conf.get<bool>(Replicator::Param::debug_log))
     {
@@ -208,19 +208,21 @@ galera::ReplicatorSMM::param_set (const std::string& key,
     if (defaults.map_.find(key) != defaults.map_.end() ||
         key                     == Param::base_host) // is my key?
     {
-        found = true;
         set_param (key, value);
-        config_.set (key, value);
+        found = true;
+        config_.set(key, value);
     }
 
-    if (key == Certification::PARAM_LOG_CONFLICTS)
-    {
-        cert_.set_log_conflicts(value);
-        return;
-    }
     // this key might be for another module
     else if (0 != key.find(common_prefix))
     {
+        try
+        {
+            cert_.param_set (key, value);
+            found = true;
+        }
+        catch (gu::NotFound&) {}
+
         try
         {
             gcs_.param_set (key, value);
@@ -234,6 +236,15 @@ galera::ReplicatorSMM::param_set (const std::string& key,
             found = true;
         }
         catch (gu::NotFound&) {}
+
+#ifdef GALERA_HAVE_SSL
+        try
+        {
+            gu::ssl_param_set (key, value, config_);
+            found = true;
+        }
+        catch (gu::NotFound&) {}
+#endif
     }
 
     if (!found) throw gu::NotFound();
