@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2023 Codership Oy <info@codership.com>
+ * Copyright (C) 2009-2019 Codership Oy <info@codership.com>
  */
 
 /*!
@@ -288,9 +288,11 @@ private:
     void populate_node_list(MessageNodeList*) const;
     void isolate(gu::datetime::Period period);
 public:
-    static std::pair<std::unique_ptr<gcomm::evs::Message>, size_t>
-    unserialize_message(const UUID&, const Datagram&);
-    void handle_msg(const Message& msg, const Datagram& dg = Datagram(),
+    static size_t unserialize_message(const UUID&,
+                                      const Datagram&,
+                                      Message*);
+    void handle_msg(const Message& msg,
+                    const Datagram& dg = Datagram(),
                     bool direct = true);
     // Protolay
     void handle_up(const void*, const Datagram&, const ProtoUpMeta&);
@@ -325,13 +327,6 @@ public:
         log_debug << self_string() << " closing in state " << state();
         if (state() != S_GATHER && state() != S_INSTALL)
         {
-            /* Leave message does not consume sequence number. Send a dummy
-             * message to trigger message acknowledgement mechanism. This is a
-             * small overhead but speeds up the leave process.*/
-            if (state() == S_OPERATIONAL) {
-                Datagram wb;
-                gu_trace(send_user(wb, 0xff, O_DROP, -1, -1));
-            }
             gu_trace(shift_to(S_LEAVING));
             gu_trace(send_leave());
             pending_leave_ = false;
