@@ -5,7 +5,7 @@ set -eu
 # $Id$
 
 # Galera library version
-VERSION="26.4.5"
+VERSION="26.4.8"
 
 get_cores()
 {
@@ -42,7 +42,9 @@ Options:
     --with-spread   configure build with spread backend (implies -c to gcs)
     --source        build source packages
     --sb            skip actual build, use the existing binaries
-    --scons         build using Scons build system (yes)
+    --cmake         build using CMake build system (yes)
+    --co            CMake options
+    --scons         build using Scons build system (no)
     --so            Sconscript option
     -j|--jobs       how many parallel jobs to use for Scons (1)
     -a|--archive    Generate source archive (tar.gz)
@@ -60,7 +62,9 @@ RELEASE=${RELEASE:-""}
 SOURCE=${SOURCE:-"no"}
 DEBUG=${DEBUG:-"no"}
 DEBUG_LEVEL=${DEBUG_LEVEL:-"0"}
-SCONS=${SCONS:-"yes"}
+CMAKE=${CMAKE:-"yes"}
+CMAKE_OPTS=${CMAKE_OPTS:-""}
+SCONS=${SCONS:-"no"}
 SCONS_OPTS=${SCONS_OPTS:-""}
 export JOBS=${JOBS:-"$(get_cores)"}
 SCRATCH=${SCRATCH:-"no"}
@@ -97,9 +101,8 @@ fi
 which dpkg >/dev/null 2>&1 && DEBIAN=${DEBIAN:-1} || DEBIAN=${DEBIAN:-0}
 
 if [ "$OS" == "FreeBSD" ]; then
-    CC=${CC:-"gcc48"}
-    CXX=${CXX:-"g++48"}
-    LD_LIBRARY_PATH=${LD_LIBRARY_PATH:-"/usr/local/lib/$(basename $CC)"}
+    CC=${CC:-"clang"}
+    CXX=${CXX:-"clang++"}
 else
     CC=${CC:-"gcc"}
     CXX=${CXX:-"g++"}
@@ -117,10 +120,12 @@ if [ -r /etc/redhat-release ]; then
   if grep 'CentOS release 5' /etc/redhat-release > /dev/null 2>&1; then
     LD_LIBRARY_PATH=/usr/local/lib64/:/usr/local/lib/:${LD_LIBRARY_PATH:-}
   fi
+  DEBIAN=0
 fi
 if [ -r /etc/issue ]; then
   if grep 'SUSE Linux Enterprise Server 11 SP1' /etc/issue > /dev/null 2>&1; then
     LD_LIBRARY_PATH=/usr/local/lib64/:/usr/local/lib/:${LD_LIBRARY_PATH:-}
+    DEBIAN=0
   fi
 fi
 
@@ -135,9 +140,9 @@ last_stage="galera"
 gainroot=""
 TARGET=${TARGET:-""} # default target
 
-while test $# -gt 0 
+while test $# -gt 0
 do
-    case $1 in 
+    case $1 in
         --stage)
             initial_stage=$2
             shift
@@ -198,8 +203,17 @@ do
         --sb)
             SKIP_BUILD="yes"
             ;;
+        --cmake)
+            CMAKE="yes"
+            SCONS="no"
+            ;;
+        --co)
+            CMAKE_OPTS="$CMAKE_OPTS $2"
+            shift
+            ;;
         --scons)
             SCONS="yes"
+            CMAKE="no"
             ;;
         --so)
             SCONS_OPTS="$SCONS_OPTS $2"
@@ -315,28 +329,20 @@ build_packages()
         DEB_BUILD_OPTIONS="version=$GALERA_VER revno=$GALERA_REV parallel=$JOBS nostrip" dpkg-buildpackage -us -uc -b -d
         RET=$?
     else
-        pushd $PKG_DIR
         if [ "$OS" == "FreeBSD" ]; then
             if test "$NO_STRIP" != "yes"; then
                 strip $build_base/{garb/garbd,libgalera_smm.so}
             fi
-            ./freebsd.sh $GALERA_VER
+            $PKG_DIR/freebsd.sh $GALERA_VER
         else # build RPM
-            ./rpm.sh $GALERA_VER
+            $PKG_DIR/rpm.sh $GALERA_VER
         fi
         RET=$?
-        popd
     fi
+    local RET=$?
 
     set -e
 
-    if [ $DEBIAN -ne 0 ]; then
-        echo Debian - Do nothing
-    elif [ "$OS" == "FreeBSD" ]; then
-        mv -f $PKG_DIR/*.tbz ./
-    else
-        mv -f $PKG_DIR/*.rpm ./
-    fi
     return $RET
 }
 
@@ -390,14 +396,49 @@ GALERA_VER=$RELEASE
 
 pushd "$build_base"
 GALERA_REV=$(git log --pretty=format:'%h' -n1.) || \
-GALERA_REV=$(bzr revno --tree -q)              || \
-GALERA_REV=$(svn info >&/dev/null && svnversion | sed s/\:/,/g) || \
-GALERA_REV="7983e"
+GALERA_REV="XXXXX"
 # trim spaces (sed is not working on Solaris, so using bash built-in)
 GALERA_REV=${GALERA_REV//[[:space:]]/}
 popd
 
-if [ "$SCONS" == "yes" ] # Build using Scons
+if [ -z "$RELEASE" ]
+then
+    source GALERA_VERSION
+    RELEASE="$GALERA_VERSION_WSREP_API.$GALERA_VERSION_MAJOR.$GALERA_VERSION_MINOR"
+fi
+if [ "$PACKAGE" == "yes" -a $DEBIAN -ne 0 ]
+then
+    # Skip extra build step when building package for debian
+    echo "Reset env for Debian package build"
+    unset CC CXX LD_LIBRARY_PATH CPPFLAGS CFLAGS CXXFLAGS
+elif [ "$CMAKE" == "yes" ] # Build using CMake
+then
+    cmake_args="$CMAKE_OPTS -DGALERA_REVISION=$GALERA_REV"
+    [ -n "$TARGET"        ] && \
+        echo "WARN: TARGET=$TARGET ignored by CMake build"
+    [ -n "$RELEASE"       ] && \
+        echo "WARN: RELEASE=$RELEASE ignored by CMake build"
+    [ "$DEBUG" == "yes"   ] && cmake_args="$cmake_args -DCMAKE_BUILD_TYPE=Debug"
+    [ -n "$EXTRA_SYSROOT" ] && \
+        echo "EXTRA_SYSROOT=$EXTRA_SYSROOT ignored by CMake build"
+
+    if [ "$SCRATCH" == "yes" ]
+    then
+        (cd $build_base && make clean) || :
+        rm -f $build_base/CMakeCache.txt
+        cmake $cmake_args $build_base
+    fi
+
+    if [ "$SKIP_BUILD" != "yes" ]
+    then
+        make -j $JOBS VERBOSE=1
+    fi
+
+    if [ $RUN_TESTS ]
+    then
+        make test ARGS=-V
+    fi
+elif [ "$SCONS" == "yes" ] # Build using Scons
 then
     # Scons variant dir, defaults to GALERA_SRC
     export SCONS_VD=$build_base
